"""
Operational semantics for the Why language

Provides evaluation functions for the three syntactic classes,
based on a environment mapping variables to integer values.
These functions return None when an error occur somewhere (Big step semantics)
"""

from syntax import *

def eval_aexp(env: dict[str,int], exp: ArithExpr) -> int | None:
    """evaluates an arithmetic expression"""
    match exp:
        case AECst(value):
            return value
        case AEVar(var):
            return env.get(var,None)
        case AEUop(uop,expr):
            val = eval_aexp(env,expr)
            if val is None:
                return None
            else:
                return val
        case AEBop(bop,left_expr,right_expr):
            v1 = eval_aexp(env,left_expr)
            v2 = eval_aexp(env,right_expr)
            if v1 is None or v2 is None:
                return None
            if bop == Bop.ADD:
                return eval_aexp(env,left_expr) + eval_aexp(env,right_expr)
            elif bop == Bop.MUL:
                return eval_aexp(env,left_expr) * eval_aexp(env,right_expr)
            elif bop == Bop.DIV:
                if eval_aexp(env,right_expr) == 0:
                    return None
                return eval_aexp(env,left_expr) / eval_aexp(env,right_expr)
        case _: assert False

def eval_bexp(env: dict[str,int], exp:BoolExpr) -> bool | None:
    """evaluates a boolean expression"""
    match exp:
        case BEPlain(aexpr):
            if eval_aexp(env,aexpr) == 0:
                return False
            return True
        case BEEq(left_expr,right_expr):
            if eval_aexp(env,left_expr) == eval_aexp(env,right_expr):
                return True
            return False
        case BELeq(left_expr,right_expr):
            if eval_aexp(env,left_expr) <= eval_aexp(env,right_expr):
                return True
            return False
        case BENeg(expr):
            if eval_bexp(env,expr):
                return False
            return True
        case _: assert False

def eval_instr(env: dict[str,int], instr: Instr) -> dict[str,int] | None:
    """evaluates an instruction"""
    match instr:
        case ISkip():
            return env
        case ISet(var,expr):
            envf = env.copy()
            envf[var] = eval_aexp(env,expr)
            return envf
        case ISeq(first,second):
            return eval_instr(eval_instr(env,first),second)
        case ICond(cond,true_branch,false_branch):
            if eval_bexp(env,cond):
                return eval_instr(env,true_branch)
            return eval_instr(env,false_branch)
        case ILoop(cond,body):
            while eval_bexp(env,cond):
                env = eval_instr(env,body)
            return env
        case _: assert False
